import os

from astropy.coordinates import SkyCoord
from astropy.table import Table, QTable
from astropy.tests.helper import assert_quantity_allclose
from astropy.io import fits
from astropy.nddata import NDData
import astropy.units as u
import numpy as np
import pytest


def _make_catalog(with_units=True, as_skycoord=False):
    """
    Create a sample catalog table with optional units and SkyCoord columns
    for testing purposes.
    """

    ra = [9.423, 9.421, 9.415] * (u.deg if with_units else 1)
    dec = [-33.711, -33.716, -33.717] * (u.deg if with_units else 1)
    obj_id = ['source1', 'souce2', 'source3']
    flux = [10, 20, 30] * (u.Jy if with_units else 1)

    tab_cls = QTable if with_units else Table

    if as_skycoord:
        sc = SkyCoord(ra, dec)
        return tab_cls(data=[sc, obj_id, flux], names=['SkyCoord', 'Obj_ID', 'flux'])
    return tab_cls(data=[ra, dec, obj_id, flux],
                   names=['RA', 'Dec', 'Obj_ID', 'flux'])


def _make_catalog_xy_radec(with_units=True):

    ra = [337.52274, 337.48273, 337.48296, 337.52333] * (u.deg if with_units else 1)
    dec = [-20.80742, -20.80741, -20.82380, -20.82425] * (u.deg if with_units else 1)
    x = [-7.27571754, 127.36624541, 126.57857853, -9.26004303]
    y = [94.28695926, 94.30831263, 35.30447687, 33.69891921]
    obj_id = ['source1', 'source2', 'source3', 'source4']
    flux = [10, 20, 30, 40] * (u.Jy if with_units else 1)

    tab_cls = QTable if with_units else Table

    return tab_cls(data=[ra, dec, x, y, obj_id, flux],
                   names=['RA', 'Dec', 'X', 'Y', 'Obj_ID', 'flux'])


def _make_catalog_string_coord_columns():
    # complelete nonsense coordinates, just care about parsing the units
    ra = ['5° 55′ 55″', '5° 55′ 55″', '5° 55′ 55″']
    dec = ['4° 44′ 44″', '4° 44′ 44″', '4° 44′ 44″']
    x = ['1', '2', '3']
    y = ['4', '5', '6']
    obj_id = ['source1', 'souce2', 'source3']
    flux = [10, 20, 30] * u.Jy

    return QTable(data=[ra, dec, x, y, obj_id, flux],
                  names=['RA', 'Dec', 'X', 'Y', 'Obj_ID', 'flux'])


def _make_catalog_no_coordinates():
    # Table to test loading catalogs that aren't 'Source Catalogs',
    col1 = ['A', 'B', 'C']
    col2 = [1, 2, 3]
    col3 = [True, False, True]

    return Table(data=[col1, col2, col3],
                 names=['col1', 'col2', 'col3'])


def test_load_catalog_no_source_positions(imviz_helper, image_2d_wcs):
    """
    A table should be able to be loaded without selecting
    an RA/Dec or X/Y pair. This table will not have the functionality
    of a 'Souce Catalog' that does have source positions
    (linking, mouseover) but it may be loaded to plot for example
    in the scatter or histrogram viewer.
    """
    imviz_helper.app.state.catalogs_in_dc = True

    catalog_obj = _make_catalog_no_coordinates()

    # load data so we can test orientation later
    data = NDData(np.ones((128, 128)), wcs=image_2d_wcs)
    imviz_helper.load(data)

    # load catalog, all columns
    imviz_helper.load(catalog_obj, col_other=['col1', 'col2', 'col3'])

    # check for the table in the data collection
    dc = imviz_helper.app.data_collection
    assert len(dc) == 2
    assert 'Catalog' in imviz_helper.app.data_collection.labels
    tab = imviz_helper.app.data_collection[1].get_object(Table)
    assert 'col1' in tab.colnames

    # make sure linking doesn't produce any errors when alingment changes.
    # this isn't relevant for this catalog with no source positons,
    # but orientation will check for the presence of certain
    # components in a table to decide not to link and we
    # want to make sure that works correctly
    imviz_helper.plugins['Orientation'].align_by = 'WCS'


def test_load_catalog_with_string_coord_cols(imviz_helper):
    """
    Test loading a catalog with string RA/Dec columns (that can be converted
    into units, e.g string representation of hourangle units) into the
    Imviz helper."""

    imviz_helper.app.state.catalogs_in_dc = True

    catalog_obj = _make_catalog_string_coord_columns()

    # load catalog
    imviz_helper.load(catalog_obj)

    dc = imviz_helper.app.data_collection
    assert len(dc) == 1
    assert 'Catalog' in imviz_helper.app.data_collection.labels

    # make coordinate columns were renamed to Right Ascension and Declination,
    # X and Y in the data collection for consistency, and that RA / Dec always
    # has units
    qtab = dc[0].get_object(QTable)
    assert 'RA' in qtab.colnames
    assert 'Dec' in qtab.colnames
    assert 'X' in qtab.colnames
    assert 'Y' in qtab.colnames
    # make sure only ra/dec/x/y/index loaded, since we didn't specify more columns
    assert len(qtab.colnames) == 5
    # and that it has the correct contents, and always has units assigned
    # when data is loaded from a unitless table, units should always be assigned
    # to the catalog in the data collection based on selections in the loader

    # go through SkyCoord to parse weird string format into deg units for comparison
    sc = SkyCoord(ra=catalog_obj['RA'], dec=catalog_obj['Dec'])

    assert_quantity_allclose(qtab['RA'], sc.ra.deg * u.deg)
    assert_quantity_allclose(qtab['Dec'], sc.dec.deg * u.deg)

    # cast data collection X/Y back to strings for comparison
    assert np.all(qtab['X'].astype(str) == catalog_obj['X'])
    assert np.all(qtab['Y'].astype(str) == catalog_obj['Y'])


@pytest.mark.parametrize("from_file", [True, False])
@pytest.mark.parametrize("with_units", [True, False])
def test_load_catalog_xy_and_radec(imviz_helper, tmp_path, from_file, with_units):

    imviz_helper.app.state.catalogs_in_dc = True

    catalog_obj = _make_catalog_xy_radec(with_units=True)

    if from_file:
        fn = os.path.join(tmp_path, "catalog.ecsv")
        catalog_obj.write(fn)
        catalog = fn
    else:
        catalog = catalog_obj

    # load catalog
    imviz_helper.load(catalog)

    dc = imviz_helper.app.data_collection
    assert len(dc) == 1
    assert 'Catalog' in imviz_helper.app.data_collection.labels

    # make sure 'RA' column was renamed to Right Ascension and 'Dec' to 'Declination'
    # in the data collection for consistency, and that the table in the data
    # collection always has units
    qtab = imviz_helper.app.data_collection[0].get_object(QTable)
    assert 'RA' in qtab.colnames
    assert 'Dec' in qtab.colnames
    assert 'X' in qtab.colnames
    assert 'Y' in qtab.colnames

    # make sure only ra/dec/x/y/index loaded, since we didn't specify more columns
    assert len(qtab.colnames) == 5
    # and that it has the correct contents
    assert_quantity_allclose(qtab['RA'], catalog_obj['RA'])
    assert_quantity_allclose(qtab['Dec'], catalog_obj['Dec'])
    assert_quantity_allclose(qtab['X'], catalog_obj['X'])
    assert_quantity_allclose(qtab['Y'], catalog_obj['Y'])


def test_import_enabled_disabled(imviz_helper):
    """
    Verify that when no coordinate column pair (RA/Dec or X/Y) is selected,
    importing is disabled, when at least one coordinate column pair is selected
    importing is enabled, and when RA is selected but Dec is not (and vice versa,
    along with the same logic for X/Y) importing is disabled.
    """

    imviz_helper.app.state.catalogs_in_dc = True

    catalog_obj = _make_catalog_xy_radec(with_units=True)

    loaders = imviz_helper.loaders
    ldr = loaders['object']
    ldr.object = catalog_obj

    ldr.format = 'Catalog'
    ldr.importer.col_ra.selected = '---'
    ldr.importer.col_dec.selected = '---'
    ldr.importer.col_x.selected = '---'
    ldr.importer.col_y.selected = '---'
    # no coordinate column pair selected, import should still be enabled
    assert ldr.importer._obj.import_disabled is False

    # when RA is selected but Dec is not, import should be disabled
    ldr.importer.col_ra.selected = 'RA'
    assert ldr.importer._obj.import_disabled is True
    # and then when Dec is selected too, import should be enabled
    ldr.importer.col_dec.selected = 'Dec'
    assert ldr.importer._obj.import_disabled is False

    # reset and test the same logic for X/Y
    ldr.importer.col_ra.selected = '---'
    ldr.importer.col_dec.selected = '---'
    ldr.importer.col_x.selected = 'X'
    ldr.importer.col_y.selected = '---'
    # now with no coordinate column pair selected, import should be disabled
    assert ldr.importer._obj.import_disabled is True
    # when Y is selected too, import should be enabled
    ldr.importer.col_y.selected = 'Y'
    assert ldr.importer._obj.import_disabled is False


@pytest.mark.parametrize("from_file", [True, False])
@pytest.mark.parametrize("with_units", [True, False])
def test_load_catalog(imviz_helper, image_2d_wcs, tmp_path, from_file, with_units):
    """
    Verify the basic functionality loading catalogs into the data collection.
    Test cases cover both in-memory Astropy tables and ECSV files, and catalogs
    with and without units on RA/Dec columns already assigned.
    """
    imviz_helper.app.state.catalogs_in_dc = True

    # basic catalog with RA, Dec, ID, and Flux columns and units
    catalog_obj = _make_catalog(with_units)

    if from_file:
        fn = os.path.join(tmp_path, "catalog.ecsv")
        catalog_obj.write(fn)
        catalog = fn
    else:
        catalog = catalog_obj

    # load image and align by WCS so loading the catalog will run through the
    # linking code to test it
    data = NDData(np.ones((128, 128)), wcs=image_2d_wcs)
    imviz_helper.load(data)
    imviz_helper.plugins['Orientation'].align_by = 'WCS'

    # load catalog
    imviz_helper.load(catalog)

    # ensure that it is in the data collection with the correct label "Catalog"
    dc = imviz_helper.app.data_collection
    assert len(dc) == 3  # image, orientation layer, and catalog
    assert 'Catalog' in imviz_helper.app.data_collection.labels

    # make sure 'RA' column was renamed to Right Ascension and 'Dec' to 'Declination'
    # in the data collection for consistency, and that the table in the data
    # collection always has units
    qtab = imviz_helper.app.data_collection[-1].get_object(QTable)
    assert 'RA' in qtab.colnames
    assert 'Dec' in qtab.colnames
    # there should also be an ID column
    assert 'ID' in qtab.colnames
    # we didn't specify a specific ID column, so it should just be the index
    # of each source
    assert np.all(qtab['ID'] == np.arange(len(catalog_obj)))
    # make sure only ra and dec (plus index) loaded, since we didn't specify more columns
    assert len(qtab.colnames) == 3
    # and that it has the correct contents, and always has units assigned
    # when data is loaded from a unitless table, units should always be assigned
    # to the catalog in the data collection based on selections in the loader
    un = 1
    if not with_units:
        un = u.deg
    assert_quantity_allclose(qtab['RA'], catalog_obj['RA'] * un)
    assert_quantity_allclose(qtab['Dec'], catalog_obj['Dec'] * un)
    assert qtab['RA'].unit == qtab['Dec'].unit == u.deg

    # make sure 'col_ra_has_unit' is True if the input catalog has units, and is
    # False if the input doesn't have units. This controls the dropdown to
    # select unit is exposed, which is eventually assigned to the table column
    # in the data collection
    ldr = imviz_helper.loaders['file' if from_file else 'object']
    setattr(ldr, 'filepath' if from_file else 'object', catalog)
    assert ldr.importer._obj.col_ra_has_unit == with_units

    # load it again, make sure label incremented by 1
    imviz_helper.load(catalog)
    assert len(dc) == 4  # image, orientation layer, and 2 catalogs
    assert 'Catalog (1)' in imviz_helper.app.data_collection.labels

    # load with custom label and check label
    imviz_helper.load(catalog, data_label='my_catalog')
    assert len(dc) == 5  # image, orientation layer, and 3 catalogs
    assert 'my_catalog' in imviz_helper.app.data_collection.labels

    # test other loader API options. switch RA and Dec col just to test
    # non-default column selection
    imviz_helper.load(catalog, data_label='with_flux', col_other='flux',
                      col_ra='Dec', col_dec='RA')
    assert len(dc) == 6  # image, orientation layer, and 4 catalogs
    assert 'flux' in dc['with_flux'].get_object(QTable).colnames
    qtab = imviz_helper.app.data_collection[-1].get_object(QTable)
    assert_quantity_allclose(qtab['RA'], catalog_obj['RA'] * un)
    assert_quantity_allclose(qtab['Dec'], catalog_obj['Dec'] * un)


@pytest.mark.parametrize("from_file", [True, False])
def test_load_catalog_skycoord(imviz_helper, tmp_path, from_file):
    """
    Test loading a catalog with a SkyCoord column into the Imviz helper.

    This test verifies that when a catalog containing a SkyCoord column is
    loaded, the loader correctly uses this column as the default for
    Right Ascension (RA) and Declination (Dec).
    """
    # test loading catalog with SkyCoord column, which should be used
    # as default for Ra and Dec

    imviz_helper.app.state.catalogs_in_dc = True

    catalog_obj = _make_catalog(as_skycoord=True)

    if from_file:
        fn = os.path.join(tmp_path, "catalog.ecsv")
        catalog_obj.write(fn)
        catalog = fn
    else:
        catalog = catalog_obj

    # load catalog
    imviz_helper.load(catalog)

    dc = imviz_helper.app.data_collection
    assert len(dc) == 1
    assert 'Catalog' in imviz_helper.app.data_collection.labels

    qtab = imviz_helper.app.data_collection[0].get_object(QTable)
    assert 'SkyCoord_RA' in qtab.colnames
    assert 'SkyCoord_Dec' in qtab.colnames
    # make sure only ra and dec (plus index) loaded, since we didn't specify more columns
    assert len(qtab.colnames) == 3
    # and that it has the correct contents, and always has units assigned
    # when data is loaded from a unitless table, units should always be assigned
    # to the catalog in the data collection based on selections in the loader
    assert_quantity_allclose(qtab['SkyCoord_RA'], catalog_obj['SkyCoord'].ra)
    assert_quantity_allclose(qtab['SkyCoord_Dec'], catalog_obj['SkyCoord'].dec)


@pytest.mark.remote_data
def test_astroquery_load_catalog_source(deconfigged_helper, catch_validate_known_exceptions):
    deconfigged_helper.app.state.catalogs_in_dc = True

    ldr = deconfigged_helper.loaders['astroquery']
    ldr.source = 'M4'
    ldr.telescope = 'Gaia'
    ldr.max_results = 10
    # TODO: remove catch_validate_known_exception
    #  when GAIA completes system maintenance (December 10, 2025 9:00 CET,
    #  this has so far proven to be a moving target...)
    # Use exception context manager to handle occasional VOTable parsing
    # errors via retrieval failures and HTTP 500 errors. Both currently due
    # to scheduled maintenance. These errors are reported as (and caught):
    # 'File does not appear to be a VOTABLE' / HTTPError: Error 500
    from astropy.io.votable.exceptions import E19
    from requests.exceptions import HTTPError
    with catch_validate_known_exceptions((E19, HTTPError, TimeoutError),
                                         stdout_text_to_check='maintenance'):
        ldr.query_archive()
    assert 'Catalog' in ldr.format.choices
    ldr.format = 'Catalog'

    ldr.importer.col_id = 'source_id'
    ldr.importer.col_other = ['parallax', 'pm', 'bp_rp', 'phot_rp_mean_mag']
    ldr.importer.viewer.create_new = 'Scatter'
    ldr.load()
    assert 'Scatter' in deconfigged_helper.viewers
    assert 'Catalog' in deconfigged_helper.viewers['Scatter'].data_menu.layer.choices


@pytest.mark.remote_data
def test_astroquery_load_catalog_from_viewer(deconfigged_helper):
    deconfigged_helper.app.state.catalogs_in_dc = True
    arr = np.ones((1489, 2048))

    # header is based on the data provided above
    hdu1 = fits.ImageHDU(arr, name='SCI')
    hdu1.header.update({'CTYPE1': 'RA---TAN',
                        'CUNIT1': 'deg',
                        'CD1_1': -7.80378407867e-05,
                        'CD1_2': 7.74904339463e-05,
                        'CRPIX1': 1025.0,
                        'CRVAL1': 6.62750450757,
                        'NAXIS1': 2048,
                        'CTYPE2': 'DEC--TAN',
                        'CUNIT2': 'deg',
                        'CD2_1': 7.74973322238e-05,
                        'CD2_2': 7.80788034973e-05,
                        'CRPIX2': 745.0,
                        'CRVAL2': 1.54470013629,
                        'NAXIS2': 1489})
    deconfigged_helper.load(hdu1, format='Image', data_label='has_wcs')

    ldr = deconfigged_helper.loaders['astroquery']
    ldr.viewer = list(deconfigged_helper.viewers.keys())[0]
    ldr.telescope = 'SDSS'
    ldr.max_results = 10
    ldr.query_archive()
    assert 'Catalog' in ldr.format.choices
    ldr.format = 'Catalog'


def test_invalid(imviz_helper, tmp_path):

    imviz_helper.app.state.catalogs_in_dc = True

    # make sure you can't load an empty table
    empty_table = Table()

    # empty table from object
    ldr = imviz_helper.loaders['object']
    ldr.object = empty_table
    assert 'Catalog' not in ldr.format.choices

    # empty table from file
    fn = os.path.join(tmp_path, "empty_catalog.ecsv")
    empty_table.write(fn)

    ldr = imviz_helper.loaders['file']
    ldr.filepath = fn
    assert 'Catalog' not in ldr.format.choices

    # while a fits file can be opened with table.read, it should not be
    # validated as a Catalog
    ldr = imviz_helper.loaders['object']
    ldr.object = fits.ImageHDU(np.ones((32, 25)))
    assert 'Image' in ldr.format.choices
    assert 'Catalog' not in ldr.format.choices


def test_scatter_viewer(deconfigged_helper):
    deconfigged_helper.app.state.catalogs_in_dc = True

    ldr = deconfigged_helper.loaders['object']
    ldr.object = _make_catalog(with_units=True)
    ldr.format = 'Catalog'

    assert 'Scatter' in ldr.importer.viewer.create_new.choices
    ldr.importer.viewer.create_new = 'Scatter'
    ldr.load()

    assert 'Scatter' in deconfigged_helper.viewers
    assert 'Scatter' in deconfigged_helper.new_viewers

    nv = deconfigged_helper.new_viewers['Scatter']
    nv.dataset = ['Catalog']
    nv.viewer_label = 'Added Scatter Viewer'
    nv()

    assert 'Added Scatter Viewer' in deconfigged_helper.viewers


def test_histogram_viewer(deconfigged_helper):
    deconfigged_helper.app.state.catalogs_in_dc = True

    ldr = deconfigged_helper.loaders['object']
    ldr.object = _make_catalog(with_units=True)
    ldr.format = 'Catalog'
    ldr.importer.col_other = ['flux']

    assert 'Histogram' in ldr.importer.viewer.create_new.choices
    ldr.importer.viewer.create_new = 'Histogram'
    ldr.load()

    assert 'Histogram' in deconfigged_helper.viewers
    assert 'Histogram' in deconfigged_helper.new_viewers

    nv = deconfigged_helper.new_viewers['Histogram']
    nv.dataset = ['Catalog']
    nv.xatt = 'flux'
    nv.viewer_label = 'Added Histogram Viewer'
    nv()

    assert 'Added Histogram Viewer' in deconfigged_helper.viewers

    po = deconfigged_helper.plugins['Plot Options']
    po.viewer = 'Added Histogram Viewer'
    po.xatt = 'RA'

    assert str(deconfigged_helper.viewers['Histogram']._obj.glue_viewer.state.x_att) == 'RA'  # noqa


def test_table_viewer(deconfigged_helper):
    deconfigged_helper.app.state.catalogs_in_dc = True

    ldr = deconfigged_helper.loaders['object']
    ldr.object = _make_catalog(with_units=True)
    ldr.format = 'Catalog'
    ldr.importer.viewer.create_new = 'Table'
    ldr.load()

    assert len(deconfigged_helper.viewers) == 1  # Table viewer created upon load
    tv = deconfigged_helper.viewers['Table']
    assert len(tv._obj.glue_viewer.layers) == 1

    # create another viewer through viewer creator
    vc = deconfigged_helper.new_viewers['Table']
    vc.dataset.select_all()
    nv = vc()

    assert len(deconfigged_helper.viewers) == 2
    assert len(nv._obj.glue_viewer.layers) == 1

    # subset creation tool should not be visible because no entries checked
    toolbar = tv._obj.glue_viewer.toolbar
    assert toolbar.tools['jdaviz:table_subset'].is_visible() is False

    tv._obj.glue_viewer.widget_table.checked = [1, 2]

    toolbar.select_tool('jdaviz:table_subset')
    assert 'Subset 1' in deconfigged_helper.plugins['Subset Tools'].subset.choices

    assert toolbar.tools['jdaviz:table_subset'].is_visible() is True
    tv._obj.glue_viewer.widget_table.checked = []
    assert toolbar.tools['jdaviz:table_subset'].is_visible() is False


def test_catalog_visibility(imviz_helper, image_2d_wcs):
    """

    # Verify that catalog visibility is toggled on/off correctly
    # based on link type and presence of pixel and/or world coordinates
    # in loaded catalog.
    """

    data = NDData(np.ones((128, 128)), wcs=image_2d_wcs)
    imviz_helper.load(data)

    # catalog with ra, dec, x, y
    orig_catalog = _make_catalog_xy_radec(with_units=True)
    # catalog with ra, dec only
    table_ra_dec_only = orig_catalog['RA', 'Dec', 'Obj_ID']
    # catalog with x, y only
    table_x_y_only = orig_catalog['X', 'Y', 'Obj_ID']

    imviz_helper.app.state.catalogs_in_dc = True
    imviz_helper.load(table_ra_dec_only,
                      data_label='catalog0')

    # since we're pixel linked and catalog has only world coordinates,
    # visiblity should be off by default
    dm = imviz_helper.viewers['imviz-0'].data_menu
    assert dm.data_labels_visible == ['Image[DATA]']

    # but if we load the catalog with X, Y, it should be visible
    imviz_helper.load(table_x_y_only,
                      data_label='catalog1')
    assert dm.data_labels_visible == ['catalog1', 'Image[DATA]']

    # now change to WCS linking
    imviz_helper.plugins['Orientation'].align_by = 'WCS'

    # load catalog with RA, Dec only again. It's default visiblity should
    # now be on since we're WCS linked
    imviz_helper.load(table_ra_dec_only,
                      data_label='catalog2')

    # the pixel-only 'catalog1' should now be hidden
    assert dm.data_labels_visible == ['catalog2', 'Image[DATA]']

    # loading a pixel-coordinate-only catalog should now be hidden by default
    # since were WCS linked
    imviz_helper.load(table_x_y_only,
                      data_label='catalog3')
    assert 'catalog3' not in dm.data_labels_visible



from numpy.testing import assert_allclose
import pytest


def test_homezoom_matchx(specviz_helper, spectrum1d):
    """
    Test HomeZoomMatchX tool activates and resets zoom in viewer.
    """
    specviz_helper.load_data(spectrum1d, data_label='test')
    viewer = specviz_helper.app.get_viewer('spectrum-viewer')

    # Change the zoom
    viewer.state.x_min = 6500
    viewer.state.x_max = 7000

    # Activate home zoom tool
    tool = viewer.toolbar.tools['jdaviz:homezoom_matchx']
    tool.activate()

    # Should reset to original limits
    assert viewer.state.x_min < 6500
    assert viewer.state.x_max > 7000


def test_boxzoom_matchx(specviz_helper, spectrum1d):
    """
    Test BoxZoomMatchX tool with zoom interaction.
    """
    specviz_helper.load_data(spectrum1d, data_label='test')
    viewer = specviz_helper.app.get_viewer('spectrum-viewer')

    # Activate box zoom tool
    tool = viewer.toolbar.tools['jdaviz:boxzoom_matchx']
    tool.activate()

    # Tool should be activated
    assert tool in viewer.toolbar.tools.values()

    # Test that the tool has the expected properties
    assert hasattr(tool, 'match_axes')
    assert tool.match_axes == ('x',)


def test_xrangezoom_matchx(specviz_helper, spectrum1d):
    """
    Test XRangeZoomMatchX tool with horizontal zoom interaction.
    """
    specviz_helper.load_data(spectrum1d, data_label='test')
    viewer = specviz_helper.app.get_viewer('spectrum-viewer')

    # Activate x-range zoom tool
    tool = viewer.toolbar.tools['jdaviz:xrangezoom_matchx']
    tool.activate()
    tool.save_prev_zoom()

    # Simulate x-range selection
    tool.interact.selected = [6500, 7000]
    tool.on_update_zoom()

    # X limits should match selection
    assert_allclose(viewer.state.x_min, 6500, rtol=1e-5)
    assert_allclose(viewer.state.x_max, 7000, rtol=1e-5)


def test_panzoom_matchx(specviz_helper, spectrum1d):
    """
    Test PanZoomMatchX tool activation and deactivation.
    """
    specviz_helper.load_data(spectrum1d, data_label='test')
    viewer = specviz_helper.app.get_viewer('spectrum-viewer')

    # Activate pan zoom tool
    tool = viewer.toolbar.tools['jdaviz:panzoom_matchx']
    tool.activate()

    # Tool should exist and be activatable
    assert tool is not None
    assert hasattr(tool, 'match_axes')

    # Deactivate
    tool.deactivate()


def test_panzoomx_matchx(specviz_helper, spectrum1d):
    """
    Test PanZoomXMatchX tool for horizontal-only panning.
    """
    specviz_helper.load_data(spectrum1d, data_label='test')
    viewer = specviz_helper.app.get_viewer('spectrum-viewer')

    # Activate pan zoom x tool
    tool = viewer.toolbar.tools['jdaviz:panzoomx_matchx']
    tool.activate()

    # Tool should exist and have correct properties
    assert tool is not None
    assert hasattr(tool, 'match_axes')
    assert tool.match_axes == ('x',)

    tool.deactivate()


def test_matched_zoom_between_viewers(specviz_helper, spectrum1d):
    """
    Test that matched zoom tools synchronize x-limits between viewers.
    """
    specviz_helper.load_data(spectrum1d, data_label='test')
    viewer = specviz_helper.app.get_viewer('spectrum-viewer')

    # Activate matched box zoom
    tool = viewer.toolbar.tools['jdaviz:boxzoom_matchx']
    tool.activate()
    tool.save_prev_zoom()

    # Change limits in one viewer
    viewer.state.x_min = 6500
    viewer.state.x_max = 7000

    # For single viewer case, just verify limits are set
    assert_allclose(viewer.state.x_min, 6500, rtol=1e-5)
    assert_allclose(viewer.state.x_max, 7000, rtol=1e-5)

    tool.deactivate()


class TestMapLimits:
    """
    Test class for _map_limits method in matched zoom tools.

    This class contains tests that directly call the _map_limits method
    to ensure coverage of its logic branches.
    """
    @staticmethod
    def _setup(deconfigged_helper, spectrum, tool_name='homezoom_matchx'):
        deconfigged_helper.load(spectrum, format='2D Spectrum')

        spectrum_viewer = deconfigged_helper.app.get_viewer('1D Spectrum')
        spectrum2d_viewer = deconfigged_helper.app.get_viewer('2D Spectrum')

        spectrum_tool = spectrum_viewer.toolbar.tools.get(f"jdaviz:{tool_name}")
        spectrum2d_tool = spectrum2d_viewer.toolbar.tools.get(f"jdaviz:{tool_name}")
        return spectrum_viewer, spectrum2d_viewer, spectrum_tool, spectrum2d_tool

    @pytest.mark.parametrize('viewer', ['Spectrum1D->Spectrum2D', 'Spectrum2D->Spectrum1D'])
    def test_map_limits_direct_call(self, deconfigged_helper, spectrum2d, viewer):
        """Test _map_limits method directly with unit conversion."""
        spectrum_viewer, spectrum2d_viewer, spectrum1d_tool, spectrum2d_tool = (
            self._setup(deconfigged_helper, spectrum2d))

        if viewer == 'Spectrum1D->Spectrum2D':
            from_viewer = spectrum_viewer
            to_viewer = spectrum2d_viewer
            tool = spectrum1d_tool
        else:
            from_viewer = spectrum2d_viewer
            to_viewer = spectrum_viewer
            tool = spectrum2d_tool

        # Change to different spectral unit
        uc = deconfigged_helper.plugins['Unit Conversion']
        uc.spectral_unit = 'nm'
        # In order for `_map_limits` to detect the spectral axis component,
        # and therefore perform the unit conversion,
        # it must have 'Wavelength' in its label.
        for component in deconfigged_helper.app.data_collection[0].components:
            if component.label == 'World 1':
                component.label = 'Wavelength'
                break

        test_limits = {'x_min': 1.0, 'x_max': 8.0}
        result = tool._map_limits(from_viewer, to_viewer, test_limits.copy())

        # Values should be the same, but units converted from um to nm
        assert_allclose(1e-3*result['x_min'], test_limits['x_min'], rtol=1e-5)
        assert_allclose(1e-3*result['x_max'], test_limits['x_max'], rtol=1e-5)

    @pytest.mark.parametrize('spec_tool', ['spectrum1d_tool', 'spectrum2d_tool'])
    def test_map_limits_by_activation(self, deconfigged_helper, spectrum2d, spec_tool):
        """
        Test that _map_limits handles unit conversions correctly from Spectrum1D to Spectrum2D.
        Also tests via activation.
        """
        spectrum_viewer, spectrum2d_viewer, spectrum1d_tool, spectrum2d_tool = (
            self._setup(deconfigged_helper, spectrum2d, 'boxzoom_matchx'))

        old_x_min, old_x_max, _, _ = spectrum_viewer.get_limits()
        if spec_tool == 'spectrum1d_tool':
            viewer = spectrum_viewer
            tool = spectrum1d_tool
            offset = 0
        else:
            viewer = spectrum2d_viewer
            tool = spectrum2d_tool
            offset = 0.5
        # _map_limits runs here
        tool.activate()

        # Change limits in spectrum viewer
        check_x_min, check_x_max, y_min, y_max = viewer.get_limits()

        # The spectrum2d tool offsets by 0.5
        assert_allclose(old_x_min, check_x_min + offset, atol=1e-5)
        assert_allclose(old_x_max, check_x_max - offset, atol=1e-5)

        # Set new limits
        tool.interact.selected = [(old_x_min + 3, y_min), (old_x_max - 3, y_max)]
        # _map_limits also runs here
        tool.on_update_zoom()

        # Limits should have changed
        new_x_min, new_x_max, _, _ = viewer.get_limits()
        assert_allclose(new_x_min, old_x_min + 3, rtol=1e-5)
        assert_allclose(new_x_max, old_x_max - 3, rtol=1e-5)

        tool.deactivate()


def test_is_matched_viewer(specviz_helper, spectrum1d):
    """
    Test the _is_matched_viewer method identifies correct viewer types.
    """
    specviz_helper.load_data(spectrum1d, data_label='test')
    viewer = specviz_helper.app.get_viewer('spectrum-viewer')

    tool = viewer.toolbar.tools['jdaviz:homezoom_matchx']

    # The tool should identify the viewer as matched
    from jdaviz.configs.specviz.plugins.viewers import (
        Spectrum1DViewer
    )
    assert isinstance(viewer, Spectrum1DViewer)
    assert tool._is_matched_viewer(viewer)


def test_matched_zoom_disable_in_other_viewer(specviz_helper, spectrum1d):
    """
    Test that activating matched zoom disables it in other viewers.
    """
    specviz_helper.load_data(spectrum1d, data_label='test')
    viewer = specviz_helper.app.get_viewer('spectrum-viewer')

    tool = viewer.toolbar.tools['jdaviz:boxzoom_matchx']

    # Check the disable flag is set
    assert tool.disable_matched_zoom_in_other_viewer is True

    # Activate the tool
    tool.activate()

    # Verify tool exists and has correct properties
    assert tool is not None
    assert hasattr(tool, '_is_matched_viewer')

    tool.deactivate()


def test_match_axes_property(specviz_helper, spectrum1d):
    """
    Test that matched zoom tools have correct match_axes property.
    """
    specviz_helper.load_data(spectrum1d, data_label='test')
    viewer = specviz_helper.app.get_viewer('spectrum-viewer')

    tool = viewer.toolbar.tools['jdaviz:homezoom_matchx']

    # Should only match x-axis for specviz
    assert tool.match_axes == ('x',)
    assert 'x_min' in tool.match_keys
    assert 'x_max' in tool.match_keys


def test_tool_icons_exist(specviz_helper, spectrum1d):
    """
    Test that all matched zoom tools have valid icon paths.
    """
    specviz_helper.load_data(spectrum1d, data_label='test')
    viewer = specviz_helper.app.get_viewer('spectrum-viewer')

    tool_ids = [
        'jdaviz:homezoom_matchx',
        'jdaviz:boxzoom_matchx',
        'jdaviz:xrangezoom_matchx',
        'jdaviz:panzoom_matchx',
        'jdaviz:panzoomx_matchx'
    ]

    for tool_id in tool_ids:
        tool = viewer.toolbar.tools[tool_id]
        assert hasattr(tool, 'icon')
        assert tool.icon is not None



import numpy as np
from astropy.nddata import NDData


def test_load_nddata(imviz_helper):
    data_a = NDData(np.random.rand(16, 16))
    data_a1 = NDData(np.random.rand(16, 16))
    data_b = NDData(np.random.rand(16, 16))
    data_b1 = NDData(np.random.rand(16, 16))
    data_b2 = NDData(np.random.rand(16, 16))
    data_c = NDData(np.random.rand(16, 16))

    imviz_helper.load(data_a, data_label='data_a')
    imviz_helper.load(data_a1, data_label='data_a1', parent='data_a[DATA]')
    imviz_helper.load(data_b, data_label='data_b')
    imviz_helper.load(data_b1, data_label='data_b1', parent='data_b[DATA]')
    imviz_helper.load(data_b2, data_label='data_b2', parent='data_b[DATA]')
    imviz_helper.load(data_c, data_label='data_c')

    dm = imviz_helper.viewers['imviz-0'].data_menu._obj
    gv = imviz_helper.viewers['imviz-0']._obj.glue_viewer

    assert ([x['zorder'] for x in dm.layer_items
             if x['label'] == 'data_c[DATA]'][0] == 6)

    # Go from 6 to 4, should redirect zorder to be 3
    # AKA below the block of data_b + children
    gv.state.layers[5].zorder = 4
    dm.layer._update_items()

    assert ([x['zorder'] for x in dm.layer_items
             if x['label'] == 'data_c[DATA]'][0] == 3)

    # Set layer with 'data_c[DATA]' to be invisible
    imviz_helper.viewers['imviz-0'].data_menu.set_layer_visibility(
        'data_c[DATA]', False)

    assert ([x['zorder'] for x in dm.layer_items
             if x['label'] == 'data_c[DATA]'][0] == 3)

    gv.state.layers[5].zorder = 1
    dm.layer._update_items()

    assert ([x['zorder'] for x in dm.layer_items
             if x['label'] == 'data_c[DATA]'][0] == 1)



from astropy.coordinates import SkyCoord
from astropy.table import Table, QTable
import astropy.units as u
import numpy as np
from traitlets import Bool, List, Unicode, observe

from jdaviz.core.loaders.importers import BaseImporterToDataCollection
from jdaviz.core.template_mixin import SelectPluginComponent
from jdaviz.core.registries import loader_importer_registry
from jdaviz.core.user_api import ImporterUserApi
from jdaviz.utils import RA_COMPS, DEC_COMPS

__all__ = ['CatalogImporter']


@loader_importer_registry("Catalog")
class CatalogImporter(BaseImporterToDataCollection):

    template_file = __file__, "./catalog.vue"

    # for catalogs with source positions in sky coordinates
    col_ra_items = List().tag(sync=True)
    col_ra_selected = Unicode().tag(sync=True)
    col_dec_items = List().tag(sync=True)
    col_dec_selected = Unicode().tag(sync=True)

    col_ra_has_unit = Bool().tag(sync=True)  # if input already has units
    col_ra_unit_items = List().tag(sync=True)
    col_ra_unit_selected = Unicode().tag(sync=True)
    col_dec_has_unit = Bool().tag(sync=True)  # if input already has units
    col_dec_unit_items = List().tag(sync=True)
    col_dec_unit_selected = Unicode().tag(sync=True)

    # for catalogs with source positions in pixel coordinates
    col_x_items = List().tag(sync=True)
    col_x_selected = Unicode().tag(sync=True)
    col_y_items = List().tag(sync=True)
    col_y_selected = Unicode().tag(sync=True)

    # (optional) to select column used as source IDs, which will be what is
    # displayed for mouseover. If None selected, an index column is used.
    col_id_items = List().tag(sync=True)
    col_id_selected = Unicode().tag(sync=True)

    # additional (optional) non-position columns to load (e.g. flux, id)
    col_other_items = List().tag(sync=True)
    col_other_selected = List().tag(sync=True)
    col_other_multiselect = Bool(True).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not self.is_valid:
            return

        # dropdowns for catalogs with source positions in sky coordinates
        self.col_ra = SelectPluginComponent(self,
                                            items='col_ra_items',
                                            selected='col_ra_selected',
                                            manual_options=self._guess_coord_cols('ra'))
        self.col_dec = SelectPluginComponent(self,
                                             items='col_dec_items',
                                             selected='col_dec_selected',
                                             manual_options=self._guess_coord_cols('dec'))

        self.col_ra_unit = SelectPluginComponent(self,
                                                 items='col_ra_unit_items',
                                                 selected='col_ra_unit_selected',
                                                 manual_options=self._valid_coord_units('ra'))
        self.col_dec_unit = SelectPluginComponent(self,
                                                  items='col_dec_unit_items',
                                                  selected='col_dec_unit_selected',
                                                  manual_options=self._valid_coord_units('dec'))

        # dropdown for source ID column
        self.col_id = SelectPluginComponent(self,
                                            items='col_id_items',
                                            selected='col_id_selected',
                                            manual_options=['Default (index)'] + self.input.colnames)  # noqa

        # dropdowns for tables with pixel source positions
        self.col_x = SelectPluginComponent(self,
                                           items='col_x_items',
                                           selected='col_x_selected',
                                           manual_options=self._guess_coord_cols('x'))
        self.col_y = SelectPluginComponent(self,
                                           items='col_y_items',
                                           selected='col_y_selected',
                                           manual_options=self._guess_coord_cols('y'))

        # dropdowns for (optional) additional columns
        self.col_other = SelectPluginComponent(self,
                                               items='col_other_items',
                                               selected='col_other_selected',
                                               manual_options=self.input.colnames,
                                               multiselect='col_other_multiselect')

    def _guess_coord_cols(self, col):
        """
        Rough guess at detecting RA/Dec/X/Y columns from input table to determine
        the initial selections for the column select dropdown. This starts
        by checking for the presence of a SkyCoord (if col is 'ra' or 'dec') or
        PixCoord column (if col is 'x' or 'y'), and next checking
        against some common source catalog column names. If no good candidate
        column is found, the initial selection in the drop down for RA/x, dec/y
        columns will be '---' (no selection)
        """

        tab = self.input
        colnames = self.input.colnames

        if colnames is None:
            return

        idx = None
        if col in ['ra', 'dec']:
            col_is_sc = [isinstance(tab[colnames[i]], SkyCoord) for i in range(len(colnames))]
            if np.any(col_is_sc):
                idx = np.where(col_is_sc)[0][0]

        if idx is None:
            # remove spaces/underscores/hyphens/quotes/parentheses and make lowercase for matching
            all_column_names = np.array([
                x.lower()
                 .replace(' ', '')
                 .replace('_', '')
                 .replace('-', '')
                 .replace('"', '')
                 .replace('(', '')
                 .replace(')', '')
                for x in colnames
            ])
            get_idx = lambda x, s, d: np.where(np.isin(x, s))[0][0] if np.any(np.isin(x, s)) else d  # noqa
            if col == 'ra':
                idx = get_idx(all_column_names, RA_COMPS, None)
            elif col == 'dec':
                idx = get_idx(all_column_names, DEC_COMPS, None)
            elif col == 'x':
                col_possibilities = ["x", "xpos", "xcentroid", "xcenter",
                                     "xpixel", "pixelx", "xpix", "ximage", "ximg",
                                     "xcoord", "xcoordinate", "sourcex", "xsource"]
                idx = get_idx(all_column_names, col_possibilities, None)
            elif col == 'y':
                col_possibilities = ["y", "ypos", "ycentroid", "ycenter",
                                     "ypixel", "pixely", "ypix", "yimage", "yimg",
                                     "ycoord", "ycoordinate", "sourcey", "ysource"]
                idx = get_idx(all_column_names, col_possibilities, None)

        # if no good candidate found, default to '---' (no selection) for
        # the default selection.
        if idx is None:
            return ['---'] + colnames
        return_cols = colnames if idx == 0 else (colnames[idx:] + colnames[:idx])
        # non-selection is the second option, so you don't have to scroll
        # all the way down to see that its an option not to load a column
        return [return_cols[0]] + ['---'] + return_cols[1:]

    def _valid_coord_units(self, coord):
        """Valid choices for Ra, Dec units."""

        choices = ['deg', 'rad', 'arcmin', 'arcsec']

        if coord == 'ra':
            choices += ['hour minute second']
        elif coord == 'dec':
            choices += ['degree arcmin arcsec']

        return choices

    @observe('col_ra_selected', 'col_dec_selected', 'col_x_selected', 'col_y_selected')
    def _on_coordinate_column_selected(self, msg):
        """
        - Check if the newly selected 'ra' or 'dec' column has units assigned
        already, to set the col_ra_has_unit and col_dec_has_unit traitlets.
        -  Make sure that ra and dec columns are not the same (unless SkyCoord) and
        disable the import button if they are the same.
        - Check if only RA or Dec is selected without the other, and disable import
        in that case.
        """

        ra = self.col_ra_selected
        dec = self.col_dec_selected
        x = self.col_x_selected
        y = self.col_y_selected

        import_disabled = False

        if msg['name'] in ('col_ra_selected', 'col_dec_selected'):

            axis = ra if msg['name'] == 'col_ra_selected' else dec

            if axis == '---':
                # no selection, assume 'has units' to disable unit selection dropdown
                if msg['name'] == 'col_ra_selected':
                    self.col_ra_has_unit = True
                elif msg['name'] == 'col_dec_selected':
                    self.col_dec_has_unit = True
                # disable import if RA is selected but Dec is not (or vice versa)
                if (ra in ['---', ''] or ra is None) != (dec in ['---', ''] or dec is None):
                    import_disabled = True
                return

            has_units = False
            if isinstance(self.input[axis], SkyCoord):
                has_units = True
            elif hasattr(self.input[axis], 'unit'):
                if self.input[axis].unit is not None:
                    has_units = True
                    # unit must be an angle unit
                    if self.input[axis].unit.physical_type != 'angle':
                        has_units = False

            # set the 'has units' traitlets for ra/dec, which determine if the unit
            # selection dropdowns should be exposed
            if msg['name'] == 'col_ra_selected':
                self.col_ra_has_unit = has_units
            elif msg['name'] == 'col_dec_selected':
                self.col_dec_has_unit = has_units

            # disable import if the same ra and dec columns are selected
            # and they are NOT a SkyCoord column (which contains both RA and Dec),
            if ra == dec and not isinstance(self.input[axis], SkyCoord):
                import_disabled = True
            else:
                import_disabled = False

            # disable import if RA is selected but Dec is not (or vice versa)
            if (ra in ['---', ''] or ra is None) != (dec in ['---', ''] or dec is None):
                import_disabled = True

        elif msg['name'] in ('col_x_selected', 'col_y_selected'):
            # disable import if RA is selected but Dec is not (or vice versa)
            if (x in ['---', ''] or x is None) != (y in ['---', ''] or y is None):
                import_disabled = True

        # finally, set the import_disabled traitlet based on what was determined
        # from the checks above
        self.import_disabled = import_disabled

    @staticmethod
    def _get_supported_viewers():
        return [{'label': 'Image', 'reference': 'imviz-image-viewer', 'allow_create': False},
                {'label': 'Scatter', 'reference': 'scatter-viewer'},
                {'label': 'Histogram', 'reference': 'histogram-viewer'},
                {'label': 'Table', 'reference': 'table-viewer'}]

    @property
    def user_api(self):
        expose = ['col_ra', 'col_dec', 'col_x', 'col_y', 'col_id', 'col_other']
        return ImporterUserApi(self, expose=expose)

    @property
    def is_valid(self):
        if self.app.state.catalogs_in_dc is False:
            return False
        if self.app.config not in ('deconfigged', 'imviz', 'mastviz'):
            # NOTE: temporary during deconfig process
            return False
        if isinstance(self.input, (Table, QTable)) and len(self.input):
            return True
        return False

    @property
    def output_cols(self):

        coordinate_cols = []
        for col in [self.col_ra_selected, self.col_dec_selected]:
            if col not in ['---', ''] and col is not None:
                coordinate_cols.append(col)
        for col in [self.col_x_selected, self.col_y_selected]:
            if col not in ['---', ''] and col is not None:
                coordinate_cols.append(col)
        if self.col_id_selected not in ['Default (index)', '', None]:
            coordinate_cols.append(self.col_id_selected)

        cols_all = coordinate_cols + self.col_other_selected

        return [col for col in set(cols_all) if col in self.input.colnames]

    @property
    def output(self):

        table = self.input[self.output_cols]
        output_table = QTable()

        # Handle RA / Dec columns, if selected
        if (self.col_ra_selected in table.colnames) and (self.col_dec_selected in table.colnames):  # noqa
            ra = self.input[self.col_ra_selected]
            dec = self.input[self.col_dec_selected]

            # The only modification made to the output table is the addition of individual
            # RA, Dec columns if the input columns are SkyCoord. This avoids unpacking
            # RA and Dec and checking if components are a SkyCoord every time they
            # are accessed.
            col_ra_selected = self.col_ra_selected  # final output col name
            col_dec_selected = self.col_dec_selected  # final output col name
            if isinstance(ra, SkyCoord):
                ra = ra.ra
                col_ra_selected = 'SkyCoord_RA'
            if isinstance(dec, SkyCoord):
                dec = dec.dec
                col_dec_selected = 'SkyCoord_Dec'

            # If the columns are strings, pass them through 'SkyCoord', which can
            # determine if the string format is recognizable as Lon/Lat coordinates.
            if isinstance(ra[0], str):
                try:
                    ra = SkyCoord(ra, ra).ra  # dummy value 'ra' twice, just to parse string
                except (ValueError, u.UnitTypeError):
                    raise ValueError("Could not parse RA column as string coordinates.")
            if isinstance(dec[0], str):
                try:
                    dec = SkyCoord(dec, dec).dec  # dummy value 'dec' twice, just to parse string
                except (ValueError, u.UnitTypeError):
                    raise ValueError("Could not parse Dec column as string coordinates.")

            # append units to RA/Dec, if they weren't loaded in with units or
            # assigned units above when parsing strings as units
            if getattr(ra, 'unit') is None:
                ra *= u.Unit(self.col_ra_unit_selected)
            if getattr(dec, 'unit') is None:
                dec *= u.Unit(self.col_dec_unit_selected)

            output_table[col_ra_selected] = ra
            output_table[col_dec_selected] = dec

            # add the selected ra/dec columns to meta
            output_table.meta['_jdaviz_loader_ra_col'] = col_ra_selected
            output_table.meta['_jdaviz_loader_dec_col'] = col_dec_selected

        # handle output construction for X and Y coordinate columns, if selected
        if (self.col_x_selected in table.colnames) and (self.col_y_selected in table.colnames):  # noqa
            # if input is a string, try to convert to floats
            if isinstance(self.input[self.col_x_selected][0], str):
                try:
                    output_table['X'] = [float(x) for x in table[self.col_x_selected]]
                except ValueError:
                    raise ValueError("Could not parse X column as numeric values.")
            if isinstance(self.input[self.col_y_selected][0], str):
                try:
                    output_table['Y'] = [float(y) for y in table[self.col_y_selected]]
                except ValueError:
                    raise ValueError("Could not parse Y column as numeric values.")

            output_table[self.col_x_selected] = table[self.col_x_selected]
            output_table[self.col_y_selected] = table[self.col_y_selected]

            # add the selected ra/dec columns to meta
            output_table.meta['_jdaviz_loader_x_col'] = self.col_x_selected
            output_table.meta['_jdaviz_loader_y_col'] = self.col_y_selected

        # add source ID column. If no column selected, just use table index
        # for now this will be added as a column named 'ID' in the output table,
        # but this should be changed to adding a component label in JDAT-5716

        if self.col_id_selected in table.colnames:
            output_table['ID'] = table[self.col_id_selected]
            output_table.meta['_jdaviz_id_col'] = self.col_id_selected
        else:
            output_table['ID'] = np.arange(len(table))
            output_table.meta['_jdaviz_id_col'] = 'ID'

        # add additional columns to output table
        for col in self.output_cols:
            if col not in output_table.colnames + [self.col_ra_selected, self.col_dec_selected]:
                output_table[col] = table[col]

        return output_table



# This file is used to configure the behavior of pytest when using the Astropy
# test infrastructure. It needs to live inside the package in order for it to
# get picked up when running the tests inside an interpreter using
# packagename.test

import os
import warnings

import numpy as np
import pytest
from astropy import units as u
from astropy.io import fits
from astropy.nddata import CCDData, StdDevUncertainty
from astropy.table import Table
from astropy.wcs import WCS
from specutils import Spectrum, SpectrumCollection, SpectrumList
from astropy.utils.masked import Masked

from jdaviz import __version__, Cubeviz, Imviz, Mosviz, Specviz, Specviz2d, Rampviz, App
from jdaviz.configs.imviz.tests.utils import (create_wfi_image_model,
                                              _image_hdu_nowcs,
                                              _image_hdu_wcs,
                                              _image_nddata_wcs)
from jdaviz.configs.imviz.plugins.parsers import HAS_ROMAN_DATAMODELS
from jdaviz.utils import NUMPY_LT_2_0
from jdaviz.core.loaders.importers.spectrum_list.spectrum_list import (
    SpectrumListImporter,
    SpectrumListConcatenatedImporter
)
from jdaviz.core.registries import loader_importer_registry
from jdaviz.core.template_mixin import PluginTemplateMixin
from jdaviz.core.registries import tray_registry

from jdaviz.pytest_memlog import (memlog_addoption, memlog_configure, memlog_runtest_setup,
                                  memlog_runtest_teardown, memlog_runtest_makereport,
                                  memlog_runtest_logreport, memlog_terminal_summary)


if not NUMPY_LT_2_0:
    np.set_printoptions(legacy="1.25")

SPECTRUM_SIZE = 10  # length of spectrum


# ============================================================================
# Memory logging plugin (memlog) - imported from pytest_memlog.py
# In CI, the memlog utilities may have already been called
# from the root conftest.py, so we avoid that by excepting
# the ValueError pytest throws.
# ============================================================================
def pytest_addoption(parser):
    """
    Register pytest options.
    """
    try:
        memlog_addoption(parser)
    except ValueError:
        pass


def pytest_runtest_setup(item):
    """
    Setup hook that records memory before test.
    """
    try:
        memlog_runtest_setup(item)
    except ValueError:
        pass


def pytest_runtest_teardown(item, nextitem):
    """
    Teardown hook that records memory after test.
    """
    try:
        memlog_runtest_teardown(item, nextitem)
    except ValueError:
        pass


# Re-export the hookwrapper directly
pytest_runtest_makereport = memlog_runtest_makereport


def pytest_runtest_logreport(report):
    """
    Log report hook that collects memory measurements from user_properties.
    """
    try:
        memlog_runtest_logreport(report)
    except ValueError:
        pass


def pytest_terminal_summary(terminalreporter, config=None):
    """
    Terminal summary hook that prints memlog summary.
    """
    try:
        memlog_terminal_summary(terminalreporter, config)
    except ValueError:
        pass


@pytest.fixture
def fake_classes_in_registries():
    """
    This fixture is meant to be used in cases where a test
    needs to check items in the registry. It provides a
    list of fake items in the various registries that could
    potentially throw off those tests if not accounted for.
    """
    return ('Test Fake Plugin',
            'Test Fake 1D Spectrum List',
            'Test Fake 1D Spectrum List Concatenated')


@tray_registry('test-fake-plugin', label='Test Fake Plugin', category='core')
class FakePlugin(PluginTemplateMixin):
    template = ''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


@loader_importer_registry('Test Fake 1D Spectrum List')
class FakeSpectrumListImporter(SpectrumListImporter):
    """A fake importer for testing/convenience purposes only.
    Mostly used to hot-update input for clean code/speed purposes.

    Usage Example:
    x = FakeSpectrumListImporter(app=deconfigged_helper.app,
                                 resolver=deconfigged_helper.loaders['object']._obj,
                                 input=premade_spectrum_list)
    """
    template = ''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.new_default_data_label = None

    @property
    def input(self):
        return super().input

    @input.setter
    def input(self, value):
        self._input = value

    @property
    def default_data_label_from_resolver(self):
        if hasattr(self, 'new_default_data_label'):
            return self.new_default_data_label
        return None


@loader_importer_registry('Test Fake 1D Spectrum List Concatenated')
class FakeSpectrumListConcatenatedImporter(SpectrumListConcatenatedImporter):
    """A fake importer for testing/convenience purposes only.
    Mostly used to hot-update input for clean code/speed purposes."""
    template = ''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.new_default_data_label = None

    @property
    def input(self):
        return super().input

    @input.setter
    def input(self, value):
        self._input = value

    @property
    def default_data_label_from_resolver(self):
        if hasattr(self, 'new_default_data_label'):
            return self.new_default_data_label
        return None


def _catch_validate_known_exceptions(exceptions_to_catch,
                                     stdout_text_to_check=''):
    """
    Context manager to catch known exceptions in CI tests. Validates the exception
    by checking for specific text in stdout. If matched, the test is skipped. If
    no text is provided, any occurrence of the exception will trigger the skip. If
    the match fails, the exception is re-raised.

    Use as:
    with _catch_known_exception(Exceptions):  # or via fixture catch_known_exceptions
        catalog_plg.search(error_on_fail=True)

    Parameters
    ----------
    exceptions_to_catch : Exception or tuple of Exceptions to catch.
    stdout_text_to_check : str, optional
        Text to match in stdout via substring matching.
        Default is '' (matches any string).
    """
    import contextlib
    import io

    @contextlib.contextmanager
    def _cm():
        buf = io.StringIO()
        try:
            with contextlib.redirect_stdout(buf):
                yield buf
        except exceptions_to_catch as etc:
            stdout_text = buf.getvalue()
            if stdout_text_to_check in stdout_text or isinstance(etc, TimeoutError):
                pytest.skip(str(etc))
            else:
                raise

    return _cm()


@pytest.fixture(scope='function')
def catch_validate_known_exceptions():
    """Context manager fixture to catch and validate known exceptions in testing."""
    return _catch_validate_known_exceptions


@pytest.fixture
def cubeviz_helper():
    return Cubeviz()


@pytest.fixture
def imviz_helper():
    return Imviz()


@pytest.fixture
def mosviz_helper():
    return Mosviz()


@pytest.fixture
def specviz_helper():
    return Specviz()


@pytest.fixture
def specviz2d_helper():
    return Specviz2d()


@pytest.fixture
def rampviz_helper():
    return Rampviz()


@pytest.fixture
def deconfigged_helper():
    return App()


@pytest.fixture
def roman_level_1_ramp():
    from roman_datamodels.datamodels import RampModel
    rng = np.random.default_rng(seed=42)

    shape = (10, 25, 25)
    data_model = RampModel.create_fake_data(shape=shape)

    data_model.data = 100 + 3 * np.cumsum(rng.uniform(size=shape), axis=0)
    return data_model


def _make_jwst_ramp(shape=(1, 10, 25, 25)):
    from stdatamodels.jwst.datamodels import Level1bModel

    rng = np.random.default_rng(seed=42)

    # JWST Level 1b ramp files have an additional preceding dimension
    # compared with Roman. This dimension is the integration number
    # in a sequence (if there's more than one in the visit).
    data_model = Level1bModel(shape)
    data_model.data = 100 + 3 * np.cumsum(rng.uniform(size=shape), axis=0)

    return data_model


@pytest.fixture
def jwst_level_1b_ramp():
    return _make_jwst_ramp()


@pytest.fixture
def jwst_level_1b_rectangular_ramp():
    return _make_jwst_ramp(shape=(1, 10, 32, 25))


@pytest.fixture
def jwst_level_2c_rate_image():
    flux_hdu = fits.ImageHDU(np.ones((32, 25)))
    flux_hdu.name = 'FLUX'
    return fits.HDUList([fits.PrimaryHDU(), flux_hdu])


@pytest.fixture
def image_2d_wcs():
    return WCS({'CTYPE1': 'RA---TAN', 'CUNIT1': 'deg', 'CDELT1': -0.0002777777778,
                'CRPIX1': 1, 'CRVAL1': 337.5202808,
                'CTYPE2': 'DEC--TAN', 'CUNIT2': 'deg', 'CDELT2': 0.0002777777778,
                'CRPIX2': 1, 'CRVAL2': -20.833333059999998})


@pytest.fixture
def sky_coord_only_source_catalog():
    """
    Create a sample source catalog with sources positioned within the
    coordinate range of the image_2d_wcs fixture.

    The catalog contains 5 sources spread across a 128x128 pixel image field,
    one roughly in the center and others towards the edges. In addition to RA
    and Dec, the catalog includes magnitude, flux, and a unique source ID.
    """

    # Create the catalog table
    catalog = Table()

    # ra and dec columns, which are required to be loaded as a source catalog
    catalog['ra'] = [337.50293, 337.52763, 337.52844, 337.47438, 337.47432] * u.deg
    catalog['dec'] = [-20.81483, -20.80438, -20.82707, -20.82683, -20.80342] * u.deg

    # additional columns
    catalog['magnitude'] = [12.5, 14.2, 13.8, 15.1, 16.3] * u.mag
    catalog['flux'] = [1.23e-12, 8.45e-13, 9.87e-13, 6.12e-13, 4.33e-13] * u.erg / (u.cm**2 * u.s)
    catalog['source_id'] = ['src_001', 'src_002', 'src_003', 'src_004', 'src_005']

    return catalog


@pytest.fixture
def spectral_cube_wcs():
    # A simple spectral cube WCS used by some tests
    wcs = WCS(naxis=3)
    wcs.wcs.ctype = 'FREQ', 'DEC--TAN', 'RA---TAN'
    wcs.wcs.set()
    return wcs


@pytest.fixture
def image_cube_hdu_obj():
    flux_hdu = fits.ImageHDU(np.ones((10, 10, 10)))
    flux_hdu.name = 'FLUX'

    mask_hdu = fits.ImageHDU(np.zeros((10, 10, 10)))
    mask_hdu.name = 'MASK'

    uncert_hdu = fits.ImageHDU(np.ones((10, 10, 10)))
    uncert_hdu.name = 'ERR'

    wcs = {
        'WCSAXES': 3, 'CRPIX1': 38.0, 'CRPIX2': 38.0, 'CRPIX3': 1.0,
        'PC1_1 ': -0.000138889, 'PC2_2 ': 0.000138889,
        'PC3_3 ': 8.33903304339E-11, 'CDELT1': 1.0, 'CDELT2': 1.0,
        'CDELT3': 1.0, 'CUNIT1': 'deg', 'CUNIT2': 'deg', 'CUNIT3': 'm',
        'CTYPE1': 'RA---TAN', 'CTYPE2': 'DEC--TAN', 'CTYPE3': 'WAVE-LOG',
        'CRVAL1': 205.4384, 'CRVAL2': 27.004754, 'CRVAL3': 3.62159598486E-07,
        'LONPOLE': 180.0, 'LATPOLE': 27.004754, 'MJDREFI': 0.0,
        'MJDREFF': 0.0, 'DATE-OBS': '2014-03-30',
        'RADESYS': 'FK5', 'EQUINOX': 2000.0
    }

    flux_hdu.header.update(wcs)
    flux_hdu.header['BUNIT'] = '1E-17 erg*s^-1*cm^-2*Angstrom^-1'

    uncert_hdu.header['BUNIT'] = '1E-17 erg*s^-1*cm^-2*Angstrom^-1'

    return fits.HDUList([fits.PrimaryHDU(), flux_hdu, uncert_hdu, mask_hdu])


@pytest.fixture
def image_cube_hdu_obj_microns():
    # Basic rectangle ramp for aperture photometry test.
    a = np.zeros((8, 9, 10)).astype(np.float32)  # (nz, ny, nx)
    for i in range(8):
        a[i, :5, :3] = i + 1
    flux_hdu = fits.ImageHDU(a)
    flux_hdu.name = 'FLUX'

    uncert_hdu = fits.ImageHDU(np.zeros((8, 9, 10)).astype(np.float32))
    uncert_hdu.name = 'ERR'

    mask_hdu = fits.ImageHDU(np.ones((8, 9, 10)).astype(np.uint16))
    mask_hdu.name = 'MASK'

    wcs = {
        'WCSAXES': 3, 'CRPIX1': 38.0, 'CRPIX2': 38.0, 'CRPIX3': 1.0,
        'CRVAL1': 205.4384, 'CRVAL2': 27.004754, 'CRVAL3': 4.890499866509344,
        'CTYPE1': 'RA---TAN', 'CTYPE2': 'DEC--TAN', 'CTYPE3': 'WAVE',
        'CUNIT1': 'deg', 'CUNIT2': 'deg', 'CUNIT3': 'um',
        'CDELT1': 3.61111097865634E-05, 'CDELT2': 3.61111097865634E-05, 'CDELT3': 0.001000000047497451,  # noqa
        'PC1_1 ': -1.0, 'PC1_2 ': 0.0, 'PC1_3 ': 0,
        'PC2_1 ': 0.0, 'PC2_2 ': 1.0, 'PC2_3 ': 0,
        'PC3_1 ': 0, 'PC3_2 ': 0, 'PC3_3 ': 1,
        'DISPAXIS': 2, 'VELOSYS': -2538.02,
        'SPECSYS': 'BARYCENT', 'RADESYS': 'ICRS', 'EQUINOX': 2000.0,
        'LONPOLE': 180.0, 'LATPOLE': 27.004754,
        'MJDREFI': 0.0, 'MJDREFF': 0.0, 'DATE-OBS': '2014-03-30'}

    flux_hdu.header.update(wcs)
    flux_hdu.header['BUNIT'] = '1E-17 erg*s^-1*cm^-2*Angstrom^-1'

    uncert_hdu.header['BUNIT'] = '1E-17 erg*s^-1*cm^-2*Angstrom^-1'

    return fits.HDUList([fits.PrimaryHDU(), flux_hdu, uncert_hdu, mask_hdu])


@pytest.fixture
def spectrum1d_cube_wcs():
    # A simple spectrum1D WCS used by some tests
    wcs = WCS(naxis=3)
    wcs.wcs.ctype = 'WAVE-LOG', 'DEC--TAN', 'RA---TAN'
    wcs.wcs.set()
    return wcs


def _create_spectrum1d_with_spectral_unit(spectralunit=u.AA, spectral_mask=None, wfss=False,
                                          exposure='0_0_0_1', source_id='0000', seed=42):
    np.random.seed(seed)

    if spectral_mask is None:
        spectral_mask = np.array([False] * SPECTRUM_SIZE)

    # We make this first so we don't have to worry about inputting different bounds
    spec_axis = Masked(np.linspace(6000, 8000, SPECTRUM_SIZE) * u.AA,
                       mask=spectral_mask)
    if spectralunit != u.AA:
        spec_axis = spec_axis.to(spectralunit)

    flux = (np.random.randn(len(spec_axis.value)) +
            10*np.exp(-0.001*(spec_axis.value-6563)**2) +
            spec_axis.value/500) * u.Jy

    uncertainty = StdDevUncertainty(np.abs(np.random.randn(len(spec_axis.value))) * u.Jy)

    meta = dict(header=dict(FILENAME="jdaviz-test-file.fits"))
    if wfss:
        meta['header'].update(dict(DATAMODL='WFSSMulti', EXPGRPID=exposure))
        meta.update(dict(source_id=source_id))

    # Note, an INFO message pops up informing the user
    # 'overwriting Masked Quantity's current mask with specified mask. [astropy.nddata.nddata]'
    # This is expected behavior albeit a nuisance. Is it possible to suppress this message?
    return Spectrum(spectral_axis=spec_axis, flux=flux, uncertainty=uncertainty,
                    mask=spectral_mask, meta=meta)


@pytest.fixture
def make_empty_spectrum():
    return Spectrum(spectral_axis=np.array([]) * u.Hz,
                    flux=np.array([]) * u.Jy,
                    uncertainty=StdDevUncertainty(np.array([])),
                    mask=np.array([]),
                    meta={})


@pytest.fixture
def spectrum1d():
    return _create_spectrum1d_with_spectral_unit()


@pytest.fixture
def partially_masked_spectrum1d():
    mask = np.array([False] * SPECTRUM_SIZE)
    mask[-3:] = True
    return _create_spectrum1d_with_spectral_unit(spectral_mask=mask)


@pytest.fixture
def wfss_spectrum1d():
    return _create_spectrum1d_with_spectral_unit(wfss=True)


# WFSS may have spectral axes that are partially masked
# and this is not allowed in specutils
@pytest.fixture
def partially_masked_wfss_spectrum1d():
    mask = np.array([False] * SPECTRUM_SIZE)
    mask[-3:] = True
    return _create_spectrum1d_with_spectral_unit(spectral_mask=mask, wfss=True, source_id='1111')


@pytest.fixture
def partially_masked_wfss_spectrum1d_exp1():
    mask = np.array([False] * SPECTRUM_SIZE)
    mask[-3:] = True
    return _create_spectrum1d_with_spectral_unit(spectral_mask=mask, wfss=True,
                                                 exposure='1_1_1', source_id='1111')


@pytest.fixture
def spectrum1d_nm():
    return _create_spectrum1d_with_spectral_unit(u.nm)


@pytest.fixture
def spectrum_collection(spectrum1d):
    sc = [spectrum1d] * 5

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        result = SpectrumCollection.from_spectra(sc)
    return result


@pytest.fixture
def premade_spectrum_list(spectrum1d, partially_masked_spectrum1d,
                          wfss_spectrum1d, partially_masked_wfss_spectrum1d,
                          partially_masked_wfss_spectrum1d_exp1):
    return SpectrumList([
        spectrum1d,
        partially_masked_spectrum1d,
        wfss_spectrum1d,
        partially_masked_wfss_spectrum1d,
        partially_masked_wfss_spectrum1d_exp1])


@pytest.fixture
def multi_order_spectrum_list(spectrum1d, spectral_orders=10):
    sc = []
    np.random.seed(42)

    for i in range(spectral_orders):

        spec_axis = (np.arange(SPECTRUM_SIZE) + 6000 + i * SPECTRUM_SIZE) * u.AA
        flux = (np.random.randn(len(spec_axis.value)) +
                10 * np.exp(-0.002 * (spec_axis.value - 6563) ** 2) +
                spec_axis.value / 500) * u.Jy
        uncertainty = StdDevUncertainty(np.abs(np.random.randn(len(spec_axis.value))) * u.Jy)
        meta = dict(header=dict(FILENAME="jdaviz-test-multi-order-file.fits"))
        spectrum1d = Spectrum(spectral_axis=spec_axis, flux=flux,
                              uncertainty=uncertainty, meta=meta)

        sc.append(spectrum1d)

    return SpectrumList(sc)


def _create_spectrum1d_cube_with_fluxunit(fluxunit=u.Jy, shape=(2, 2, 4), with_uncerts=False):
    # nz=2 nx=2 ny=4
    flux = np.arange(np.prod(shape)).reshape(shape) * fluxunit
    wcs_dict = {"CTYPE1": "RA---TAN", "CTYPE2": "DEC--TAN", "CTYPE3": "WAVE-LOG",
                "CRVAL1": 205, "CRVAL2": 27, "CRVAL3": 4.622e-7,
                "CDELT1": -0.0001, "CDELT2": 0.0001, "CDELT3": 8e-11,
                "CRPIX1": 0, "CRPIX2": 0, "CRPIX3": 0, "PIXAR_SR": 10.,
                # Need these for aperture photometry test.
                "TELESCOP": "JWST", "BUNIT": fluxunit.to_string(), "PIXAR_A2": 0.01}
    w = WCS(wcs_dict)
    if with_uncerts:
        uncert = StdDevUncertainty(np.abs(np.random.normal(flux) * fluxunit))

        return Spectrum(flux=flux,
                        uncertainty=uncert,
                        wcs=w,
                        meta=wcs_dict)
    else:
        return Spectrum(flux=flux, wcs=w, meta=wcs_dict)


@pytest.fixture
def spectrum1d_cube():
    return _create_spectrum1d_cube_with_fluxunit(fluxunit=u.Jy)


@pytest.fixture
def spectrum1d_cube_with_uncerts():
    return _create_spectrum1d_cube_with_fluxunit(fluxunit=u.Jy, with_uncerts=True)


@pytest.fixture
def spectrum1d_cube_larger():
    return _create_spectrum1d_cube_with_fluxunit(fluxunit=u.Jy, shape=(SPECTRUM_SIZE, 2, 4))


@pytest.fixture
def spectrum1d_cube_largest():
    wcs_dict = {"CTYPE1": "WAVE-LOG", "CTYPE2": "DEC--TAN", "CTYPE3": "RA---TAN",
                "CRVAL1": 4.622e-7, "CRVAL2": 27, "CRVAL3": 205,
                "CDELT1": 8e-11, "CDELT2": 0.0001, "CDELT3": -0.0001,
                "CRPIX1": 0, "CRPIX2": 0, "CRPIX3": 0}
    w = WCS(wcs_dict)
    flux = np.zeros((20, 30, 3001), dtype=np.float32)  # nx=20 ny=30 nz=3001
    flux[1:11, 5:15, :] = 1  # Bright corner
    return Spectrum(flux=flux * u.Jy, wcs=w, meta=wcs_dict)


@pytest.fixture
def spectrum1d_cube_custom_fluxunit():
    return _create_spectrum1d_cube_with_fluxunit


@pytest.fixture
def spectrum1d_cube_fluxunit_jy_per_steradian():
    return _create_spectrum1d_cube_with_fluxunit(fluxunit=u.Jy/u.sr, shape=(10, 4, 5),
                                                 with_uncerts=True)


@pytest.fixture
def spectrum1d_cube_sb_unit():
    # similar fixture to spectrum1d_cube_fluxunit_jy_per_steradian, but no uncerts
    # and different shape. can probably remove one of these eventually
    return _create_spectrum1d_cube_with_fluxunit(fluxunit=u.Jy / u.sr)


@pytest.fixture
def mos_spectrum1d(mos_spectrum2d):
    '''
    A specially defined Spectrum1d that matches the corresponding spectrum2d below.

    TODO: this fixture should be replaced by the global spectrum1d fixture defined in
    jdaviz/conftest.py AFTER reforming the spectrum2d fixture below to match the
    global spectrum1d fixture.

    Unless linking the two is required, try to use the global spectrum1d fixture.
    '''
    spec_axis = mos_spectrum2d.spectral_axis
    np.random.seed(42)
    flux = (np.random.randn(len(spec_axis.value)) +
            10*np.exp(-0.001*(spec_axis.value-6563)**2) +
            spec_axis.value/500) * u.Jy

    return Spectrum(spectral_axis=spec_axis, flux=flux)


@pytest.fixture
def spectrum2d():
    '''
    A simple 2D Spectrum with a center "trace" array rising from 0 to 10
    with two "zero array" buffers above and below
    '''
    data = np.zeros((5, 10))
    data[3] = np.arange(10)

    return Spectrum(flux=data*u.MJy, spectral_axis=data[3]*u.um)


def _generate_mos_spectrum2d():
    header = {
        'WCSAXES': 2,
        'CRPIX1': 0.0, 'CRPIX2': 1024.5,
        'CDELT1': 1E-06, 'CDELT2': 2.9256727777778E-05,
        'CUNIT1': 'm', 'CUNIT2': 'deg',
        'CTYPE1': 'WAVE', 'CTYPE2': 'OFFSET',
        'CRVAL1': 0.0, 'CRVAL2': 5.0,
        'RADESYS': 'ICRS', 'SPECSYS': 'BARYCENT'}
    np.random.seed(42)
    data = np.random.sample((15, 1024)) * u.Jy
    return data, header


@pytest.fixture
def mos_spectrum2d():
    '''
    A specially defined 2D (spatial) Spectrum whose wavelength range matches the
    mos-specific 1D spectrum.

    TODO: This should be reformed to match the global Spectrum defined above so that we may
    deprecate the mos-specific spectrum1d.
    '''
    data, header = _generate_mos_spectrum2d()
    wcs = WCS(header)
    return Spectrum(data, wcs=wcs, meta=header)


@pytest.fixture
def mos_spectrum2d_as_hdulist():
    data, header = _generate_mos_spectrum2d()
    hdu = fits.ImageHDU(data.value)
    hdu.header.update(header)

    # This layout is to trick specutils to think it is JWST s2d
    hdulist = fits.HDUList([fits.PrimaryHDU(), hdu, hdu])
    hdulist[0].header["TELESCOP"] = "JWST"
    hdulist[1].name = "SCI"
    hdulist[1].ver = 1
    hdulist[2].name = "SCI"
    hdulist[2].ver = 2

    return hdulist


@pytest.fixture
def mos_image():
    header = {
        'WCSAXES': 2,
        'CRPIX1': 937.0, 'CRPIX2': 696.0,
        'CDELT1': -1.5182221158397e-05, 'CDELT2': 1.5182221158397e-05,
        'CUNIT1': 'deg', 'CUNIT2': 'deg',
        'CTYPE1': 'RA---TAN', 'CTYPE2': 'DEC--TAN',
        'CRVAL1': 5.0155198140981, 'CRVAL2': 5.002450989248,
        'LONPOLE': 180.0, 'LATPOLE': 5.002450989248,
        'DATEREF': '1858-11-17', 'MJDREFI': 0.0, 'MJDREFF': 0.0,
        'RADESYS': 'ICRS'}
    wcs = WCS(header)
    np.random.seed(42)
    data = np.random.sample((55, 55))
    return CCDData(data, wcs=wcs, unit='Jy', meta=header)


@pytest.fixture
def roman_imagemodel():
    if HAS_ROMAN_DATAMODELS:
        return create_wfi_image_model((20, 10))


@pytest.fixture
def image_hdu_nowcs():
    return _image_hdu_nowcs()


@pytest.fixture
def image_hdu_wcs():
    return _image_hdu_wcs()


@pytest.fixture
def multi_extension_image_hdu_wcs():
    return fits.HDUList([fits.PrimaryHDU(),
                         _image_hdu_wcs(),
                         _image_hdu_nowcs(np.zeros((10, 10)), name='MASK'),
                         _image_hdu_nowcs(name='ERR'),
                         _image_hdu_nowcs(name='DQ')])


@pytest.fixture
def image_nddata_wcs():
    return _image_nddata_wcs()


# Copied over from https://github.com/spacetelescope/ci_watson
@pytest.fixture(scope='function')
def _jail(tmp_path):
    """Perform test in a pristine temporary working directory."""
    old_dir = os.getcwd()
    os.chdir(tmp_path)
    try:
        yield str(tmp_path)
    finally:
        os.chdir(old_dir)


try:
    from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS
except ImportError:
    PYTEST_HEADER_MODULES = {}
    TESTED_VERSIONS = {}


def pytest_configure(config):
    """
    Configure pytest, including memlog initialization.
    """
    # Initialize memlog
    memlog_configure(config)

    # Configure pytest header modules
    PYTEST_HEADER_MODULES['astropy'] = 'astropy'
    PYTEST_HEADER_MODULES['pyyaml'] = 'yaml'
    PYTEST_HEADER_MODULES['scikit-image'] = 'skimage'
    PYTEST_HEADER_MODULES['specutils'] = 'specutils'
    PYTEST_HEADER_MODULES['specreduce'] = 'specreduce'
    PYTEST_HEADER_MODULES['asteval'] = 'asteval'
    PYTEST_HEADER_MODULES['echo'] = 'echo'
    PYTEST_HEADER_MODULES['idna'] = 'idna'
    PYTEST_HEADER_MODULES['traitlets'] = 'traitlets'
    PYTEST_HEADER_MODULES['bqplot'] = 'bqplot'
    PYTEST_HEADER_MODULES['bqplot-image-gl'] = 'bqplot_image_gl'
    PYTEST_HEADER_MODULES['glue-core'] = 'glue'
    PYTEST_HEADER_MODULES['glue-jupyter'] = 'glue_jupyter'
    PYTEST_HEADER_MODULES['glue-astronomy'] = 'glue_astronomy'
    PYTEST_HEADER_MODULES['ipyvue'] = 'ipyvue'
    PYTEST_HEADER_MODULES['ipyvuetify'] = 'ipyvuetify'
    PYTEST_HEADER_MODULES['ipysplitpanes'] = 'ipysplitpanes'
    PYTEST_HEADER_MODULES['ipygoldenlayout'] = 'ipygoldenlayout'
    PYTEST_HEADER_MODULES['ipypopout'] = 'ipypopout'
    PYTEST_HEADER_MODULES['solara'] = 'solara'
    PYTEST_HEADER_MODULES['vispy'] = 'vispy'
    PYTEST_HEADER_MODULES['gwcs'] = 'gwcs'
    PYTEST_HEADER_MODULES['asdf'] = 'asdf'
    PYTEST_HEADER_MODULES['stdatamodels'] = 'stdatamodels'
    PYTEST_HEADER_MODULES['roman_datamodels'] = 'roman_datamodels'

    TESTED_VERSIONS['jdaviz'] = __version__



from jdaviz.configs.cubeviz.plugins.viewers import CubevizImageView
from jdaviz.core.viewer_creators import BaseViewerCreator
from jdaviz.core.registries import viewer_creator_registry


__all__ = ['Spectrum3DViewerCreator']


@viewer_creator_registry('3D Spectrum', overwrite=True)
class Spectrum3DViewerCreator(BaseViewerCreator):
    template_file = __file__, "../base_viewer_creator.vue"

    def __init__(self, app, **kwargs):
        super().__init__(app, **kwargs)
        self.dataset.filters = ['is_flux_cube']
        self.viewer_label_default = '3D Spectrum'

    @property
    def viewer_class(self):
        return CubevizImageView



from astropy.io import fits
from io import BytesIO
import numpy as np
import pytest
import warnings

from pyvo.io.vosi.endpoint import parse_capabilities
from pyvo.utils.xml.exceptions import UnknownElementWarning

from jdaviz.configs.imviz.tests.utils import BaseImviz_WCS_WCS


class fake_siaresult:
    """A mock class that simulates a SIAResult"""

    def __init__(self, attrs):
        for attr, value in attrs.items():
            self.__setattr__(attr, value)

    def getdataurl(self):
        return "Fake URL"


# TODO: Update all _obj calls to formal API calls once Plugin API is available
class TestVOImvizLocal(BaseImviz_WCS_WCS):
    _data_center_coords = {
        "has_wcs_1[SCI,1]": {"ra": 337.51894336761296, "dec": -20.832083054811765},
        "has_wcs_2[SCI,1]": {"ra": 337.51924057481, "dec": -20.83208305686149},
    }

    def test_autocenter_coords(self):
        """
        Loads two data products and checks the plugin correctly adjusts the automatically-derived
        coordinates of the center of the viewer when the viewer dropdown is changed.

        Also verify changing autocoord to a blank viewer with no data properly empties the source
        field.
        """
        # Create a second viewer and remove second dataset from first viewer to avoid ambiguity
        self.imviz.create_image_viewer()
        self.imviz.app.remove_data_from_viewer("imviz-0", "has_wcs_2[SCI,1]")

        # Check default viewer is "Manual"
        vo_ldr = self.imviz.loaders["virtual observatory"]._obj
        assert vo_ldr.viewer.selected == "Manual"
        assert vo_ldr.source == ""

        # Switch to first viewer and verify coordinates have switched
        vo_ldr.viewer.selected = "imviz-0"
        ra_str, dec_str = vo_ldr.source.split()
        np.testing.assert_allclose(
            float(ra_str), self._data_center_coords["has_wcs_1[SCI,1]"]["ra"]
        )
        np.testing.assert_allclose(
            float(dec_str), self._data_center_coords["has_wcs_1[SCI,1]"]["dec"]
        )

        # Switch to second viewer without data and verify autocoord gracefully clears source field
        vo_ldr.viewer.selected = "imviz-1"
        assert vo_ldr.source == ""

        # Now load second data into second viewer and verify coordinates
        self.imviz.app.add_data_to_viewer("imviz-1", "has_wcs_2[SCI,1]")
        ra_str, dec_str = vo_ldr.source.split()
        np.testing.assert_allclose(
            float(ra_str), self._data_center_coords["has_wcs_2[SCI,1]"]["ra"]
        )
        np.testing.assert_allclose(
            float(dec_str), self._data_center_coords["has_wcs_2[SCI,1]"]["dec"]
        )


def test_link_type_autocoord(imviz_helper):
    """
    Tests switching linking types forces recalculation of viewer center coordinates
    """
    # First data with WCS, same as the one in BaseImviz_WCS_NoWCS.
    hdu1 = fits.ImageHDU(np.random.rand(100, 100), name="SCI")
    hdu1.header.update(
        {
            "CTYPE1": "RA---TAN",
            "CUNIT1": "deg",
            "CDELT1": -2.777777778,
            "CRPIX1": 1,
            "CRVAL1": 337.5202808,
            "NAXIS1": 10,
            "CTYPE2": "DEC--TAN",
            "CUNIT2": "deg",
            "CDELT2": 2.777777778,
            "CRPIX2": 1,
            "CRVAL2": -20.833333059999998,
            "NAXIS2": 10,
        }
    )
    imviz_helper.load_data(hdu1, data_label="has_wcs_1")

    # Second data with WCS, similar to above but dithered by 1 pixel in X.
    hdu2 = fits.ImageHDU(np.ones((10, 10)), name="SCI")
    hdu2.header.update(
        {
            "CTYPE1": "RA---TAN",
            "CUNIT1": "deg",
            "CDELT1": -0.0002777777778,
            "CRPIX1": 2,
            "CRVAL1": 137.5202808,
            "NAXIS1": 10,
            "CTYPE2": "DEC--TAN",
            "CUNIT2": "deg",
            "CDELT2": 0.0002777777778,
            "CRPIX2": 1,
            "CRVAL2": -20.833333059999998,
            "NAXIS2": 10,
        }
    )
    imviz_helper.load_data(hdu2, data_label="has_wcs_2")

    vo_ldr = imviz_helper.loaders["virtual observatory"]._obj
    vo_ldr.viewer.selected = "imviz-0"
    vo_ldr.center_on_data()
    ra_str, dec_str = vo_ldr.source.split()
    np.testing.assert_allclose(float(ra_str), 284.2101962057667)
    np.testing.assert_allclose(float(dec_str), 32.23616603681311)

    imviz_helper.plugins["Orientation"].align_by = "WCS"

    ra_str, dec_str = vo_ldr.source.split()

    # Large absolute tolerances due to WCS center coordinate bug (see issue 3225)
    # Truth values may need to be reevaluated
    np.testing.assert_allclose(float(ra_str), 239.18585, atol=30)
    np.testing.assert_allclose(float(dec_str), -9.905948925234416, atol=30)


class TestVOXMLInjectionWarning:
    """
    Test class for VO XML Injection warning scenarios.

    This class contains tests that demonstrate the behavior of
    UnknownElementWarning when parsing XML with non-standard elements.
    """

    @pytest.fixture(autouse=True)
    def setup_method(self):
        """
        Setup method to initialize common XML data for tests.
        """
        # XML with non-standard <limits> element
        self.xml_with_limits = b"""<?xml version="1.0" encoding="UTF-8"?>
            <capabilities xmlns="http://www.ivoa.net/xml/VOSICapabilities/v1.0"
                          xmlns:vr="http://www.ivoa.net/xml/VOResource/v1.0"
                          xmlns:tr="http://www.ivoa.net/xml/TAPRegExt/v1.0"
                          xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
              <capability standardID="ivo://ivoa.net/std/TAP" xsi:type="tr:TableAccess">
                <interface xsi:type="vr:WebService" role="std">
                  <accessURL use="base">http://example.com/tap</accessURL>
                </interface>
                <language>
                  <name>ADQL</name>
                  <version ivo-id="ivo://ivoa.net/std/ADQL#v2.0">2.0</version>
                </language>
                <outputFormat>
                  <mime>application/x-votable+xml</mime>
                </outputFormat>
                <limits>
                  <default>
                    <executionDuration>3600</executionDuration>
                    <outputLimit unit="row">10000</outputLimit>
                  </default>
                </limits>
              </capability>
            </capabilities>"""

    def test_direct_xml_parsing_triggers_warning(self):
        """Parse XML with <limits> and check the warning."""
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            result = parse_capabilities(BytesIO(self.xml_with_limits))

            # Verify the warning was triggered
            limits_warnings = [
                warning for warning in w
                if issubclass(warning.category, UnknownElementWarning)
                and 'limits' in str(warning.message).lower()
            ]

            # Assert we got the warning
            assert len(limits_warnings) == 1
            assert limits_warnings[0].category == UnknownElementWarning
            assert "Unknown element limits" in str(limits_warnings[0].message)

            # Assert parsing still succeeded
            assert len(result) > 0

    @pytest.mark.filterwarnings(
        "ignore::pyvo.utils.xml.exceptions.UnknownElementWarning"
    )
    def test_xml_parsing_with_filter_passes(self):
        """
        Parse XML with <limits> WITH the warning filter decorator.

        This test should pass because the decorator filters the warning
        UnknownElementWarning, just like it sometimes does in test_coverage_toggle.
        """
        # This should NOT fail because the decorator filters the warning
        result = parse_capabilities(BytesIO(self.xml_with_limits))

        # Verify parsing succeeded and we got a result
        assert len(result) > 0


@pytest.mark.remote_data
class TestVOImvizRemote:

    def _init_vo_ldr_M51(self, imviz_helper):
        """
        Initialize vo loader with common test parameters

        Parameters
        ----------
        imviz_helper : `~jdaviz.configs.imviz.helper`
            Instance of Imviz in which to initialize the VO loader

        Returns
        -------
        vo_ldr_api : VO loader user API instance
        """
        vo_ldr = imviz_helper.loaders["virtual observatory"]

        # Sets common args for Remote Testing
        vo_ldr.viewer.selected = "Manual"
        vo_ldr.source = "M51"
        vo_ldr.radius = 1
        vo_ldr.radius_unit.selected = "deg"
        vo_ldr.waveband.selected = "optical"

        return vo_ldr

    def test_query_registry_args(self, imviz_helper):
        """Ensure we don't query registry if we're missing required arguments"""
        # If waveband isn't selected, plugin should ignore our registry query attempts
        vo_ldr = imviz_helper.loaders["virtual observatory"]
        vo_ldr.waveband.selected = ""
        assert len(vo_ldr.resource.choices) == 0

        # If waveband selected and coverage filtering, can't query registry if don't have a source
        vo_ldr.resource_filter_coverage = True
        vo_ldr.source = ""
        with pytest.raises(
            ValueError,
            match="Source is required for registry querying when coverage filtering is enabled.",
        ):
            # Setting the waveband from nothing to something will trigger the query
            vo_ldr.waveband.selected = "optical"
            # Also verify we get a snackbar message for it
            assert any(
                "Source is required" in d["text"]
                for d in imviz_helper.plugins['Logger'].history
            )

        # If waveband selected, but NOT filtering by coverage, then allow registry query
        vo_ldr.resource_filter_coverage = False
        assert len(vo_ldr.resource.choices) > 0

    @pytest.mark.skip(reason="need to investigate failure from upstream change")
    @pytest.mark.filterwarnings(
        "ignore::pyvo.utils.xml.exceptions.UnknownElementWarning"
    )
    def test_coverage_toggle(self, imviz_helper):
        """
        Test that disabling the coverage toggle returns more available services

        NOTE: This does assume there exists at least one survey that does NOT report coverage
        within a 1-degree circle around the above-defined source position. Otherwise, returned
        resource lists will be identical.
        """
        # Set Common Args
        vo_ldr = self._init_vo_ldr_M51(imviz_helper)

        # Retrieve registry options with filtering on
        vo_ldr.resource_filter_coverage = True
        assert vo_ldr._obj.resources_loading is False
        filtered_resources = vo_ldr.resource.choices
        assert len(filtered_resources) > 0

        # Retrieve registry options with filtering off
        vo_ldr.resource_filter_coverage = False
        assert vo_ldr._obj.resources_loading is False
        nonfiltered_resources = vo_ldr.resource.choices
        # Even if the warning is triggered, this line should still pass
        # because the execution should still continue. If it doesn't,
        # then we know the warning solution did not work.
        assert len(nonfiltered_resources) > 0

        # Nonfiltered resources should be more than filtered resources
        assert len(nonfiltered_resources) > len(filtered_resources)

    def test_target_lookup_warnings(self, imviz_helper):
        """
        Tests that appropriate errors and guardrails protect the user
        when a provided source is irresolvable
        """
        # Manually set the source to a fake target
        vo_ldr = imviz_helper.loaders["virtual observatory"]
        vo_ldr.source = "ThisIsAFakeTargetThatWontResolveToAnything"
        vo_ldr.radius = 1
        vo_ldr.radius_unit.selected = "deg"

        # If we have coverage filtering on, we should get an error
        vo_ldr.resource_filter_coverage = True
        expected_error_msg = "Unable to resolve source coordinates"
        with pytest.raises(LookupError, match=expected_error_msg):
            vo_ldr.waveband.selected = "optical"
            assert any(
                expected_error_msg in d["text"]
                for d in imviz_helper.plugins['Logger'].history
            )
            assert len(vo_ldr.resource.choices) == 0

        # By clearing coverage filtering, we should now be able to query the registry
        # and return the full list of available resources:
        vo_ldr.resource_filter_coverage = False
        assert len(vo_ldr.resource.choices) > 0

        # However, if we try to query a resource, we should be prevented
        # since the source still isn't resolvable.
        # Clear existing messages
        imviz_helper.plugins['Logger'].clear_history()
        vo_ldr.resource.selected = "HST.M51"
        vo_ldr.query_archive()
        assert any(
            expected_error_msg in d["text"]
            for d in imviz_helper.plugins['Logger'].history
        )

    @pytest.mark.filterwarnings("ignore:Some non-standard WCS keywords were excluded")
    def test_HSTM51_data_url(self, imviz_helper):
        vo_ldr = self._init_vo_ldr_M51(imviz_helper)

        # Select HST.M51 survey
        # Coverage not implemented for HST.M51
        vo_ldr.resource_filter_coverage = False
        assert "HST.M51" in vo_ldr.resource.choices
        vo_ldr.resource.selected = "HST.M51"
        vo_ldr.query_archive()
        assert len(vo_ldr.file_table._obj.items) > 0

        # Load first data product
        assert vo_ldr._obj.get_selected_url() is None
        vo_ldr.file_table.select_rows(0)
        assert vo_ldr._obj.get_selected_url() is not None and len(vo_ldr._obj.get_selected_url()) > 0  # noqa
